"""

    routr.utilities -- utility code
    ===============================

"""

import sys

__all__ = ("import_string", "cached_property", "ImportStringError", "join")

class cached_property(object):
    """ Just like ``property`` but computed only once"""

    def __init__(self, func):
        self.func = func
        self.__name__ = func.__name__
        self.__doc__ = func.__doc__

    def __get__(self, obj, cls):
        if obj is None:
            return self
        val = self.func(obj)
        obj.__dict__[self.__name__] = val
        return val

def import_string(import_name, silent=False):
    """Imports an object based on a string.  This is useful if you want to
    use import paths as endpoints or something similar.  An import path can
    be specified either in dotted notation (``xml.sax.saxutils.escape``)
    or with a colon as object delimiter (``xml.sax.saxutils:escape``).

    If `silent` is True the return value will be `None` if the import fails.

    For better debugging we recommend the new :func:`import_module`
    function to be used instead.

    :param import_name: the dotted name for the object to import.
    :param silent: if set to `True` import errors are ignored and
                   `None` is returned instead.
    :return: imported object

    :copyright: (c) 2011 by the Werkzeug Team
    """
    # force the import name to automatically convert to strings
    if isinstance(import_name, unicode):
        import_name = str(import_name)
    try:
        if ':' in import_name:
            module, obj = import_name.split(':', 1)
        elif '.' in import_name:
            module, obj = import_name.rsplit('.', 1)
        else:
            return __import__(import_name)
        # __import__ is not able to handle unicode strings in the fromlist
        # if the module is a package
        if isinstance(obj, unicode):
            obj = obj.encode('utf-8')
        try:
            return getattr(__import__(module, None, None, [obj]), obj)
        except (ImportError, AttributeError):
            # support importing modules not yet set up by the parent module
            # (or package for that matter)
            modname = module + '.' + obj
            __import__(modname)
            return sys.modules[modname]
    except ImportError, e:
        if not silent:
            raise ImportStringError(import_name, e), None, sys.exc_info()[2]

class ImportStringError(ImportError):
    """Provides information about a failed :func:`import_string` attempt.

    :copyright: (c) 2011 by the Werkzeug Team
    """

    #: String in dotted notation that failed to be imported.
    import_name = None
    #: Wrapped exception.
    exception = None

    def __init__(self, import_name, exception):
        self.import_name = import_name
        self.exception = exception

        msg = (
            'import_string() failed for %r. Possible reasons are:\n\n'
            '- missing __init__.py in a package;\n'
            '- package or module path not included in sys.path;\n'
            '- duplicated package or module name taking precedence in '
            'sys.path;\n'
            '- missing module, class, function or variable;\n\n'
            'Debugged import:\n\n%s\n\n'
            'Original exception:\n\n%s: %s')

        name = ''
        tracked = []
        for part in import_name.replace(':', '.').split('.'):
            name += (name and '.') + part
            imported = import_string(name, silent=True)
            if imported:
                tracked.append((name, imported.__file__))
            else:
                track = ['- %r found in %r.' % (n, i) for n, i in tracked]
                track.append('- %r not found.' % name)
                msg = msg % (import_name, '\n'.join(track),
                             exception.__class__.__name__, str(exception))
                break

        ImportError.__init__(self, msg)

    def __repr__(self):
        return '<%s(%r, %r)>' % (self.__class__.__name__, self.import_name,
                                 self.exception)

def join(a, b):
    a = a or ""
    b = b or ""
    return a.rstrip("/") + "/" + b.lstrip("/")



"""

    routr -- define routes
    ======================

"""

import re
from urllib import urlencode
from pkg_resources import iter_entry_points

from webob.exc import HTTPException
from routr.utils import import_string, cached_property, join
from routr.exc import (
    NoMatchFound, NoURLPatternMatched, RouteGuarded,
    MethodNotAllowed, RouteConfigurationError, InvalidRoutePattern,
    RouteReversalError)

__all__ = (
    "Configuration", "route", "include", "plug", "Trace",
    "Route", "Endpoint", "RootEndpoint", "RouteGroup", "HTTPMethod",
    "GET", "POST", "PUT", "DELETE", "HEAD", "OPTIONS", "TRACE", "PATCH",
    "NoMatchFound", "RouteConfigurationError")

class HTTPMethod(str):
    """ HTTP method"""

GET     = HTTPMethod("GET")
POST    = HTTPMethod("POST")
PUT     = HTTPMethod("PUT")
DELETE  = HTTPMethod("DELETE")
HEAD    = HTTPMethod("HEAD")
OPTIONS = HTTPMethod("OPTIONS")
TRACE   = HTTPMethod("TRACE")
PATCH   = HTTPMethod("PATCH")

class Trace(object):
    """ Result of route matching

    Represents a trace of matched routes towards the last one called endpoint.

    :attr args:
        collected positional arguments
    :attr kwargs:
        collected keyword arguments
    :attr routes:
        a list of matched routes with the endpoint route being the last one
    :attr endpoint:
        matched endpoint route

    Attributes of :attr:`.endpoint` route are made available on trace itself.
    """

    def __init__(self, args, kwargs, routes, payload=None):
        self.__dict__["args"] = args
        self.__dict__["kwargs"] = kwargs
        self.__dict__["routes"] = routes
        self.__dict__["payload"] = payload or {}

    @property
    def endpoint(self):
        return self.routes[-1] if self.routes else None

    @property
    def target(self):
        return self.endpoint.target

    def __add__(self, tr):
        args = self.args + tr.args
        kwargs = dict(self.kwargs)
        kwargs.update(tr.kwargs)
        routes = list(self.routes)
        routes.extend(tr.routes)
        payload = dict(self.payload)
        payload.update(tr.payload)
        return self.__class__(args, kwargs, routes, payload)

    def __setattr__(self, name, value):
        self.payload[name] = value

    def __getattr__(self, name):
        if not name in self.payload:
            raise AttributeError(name)
        return self.payload[name]

class Route(object):
    """ Base class for routes

    :param cfg:
        configuration object
    :param guards:
        a list of guards
    :param pattern:
        pattern for URL pattern
    """

    def __init__(self, cfg, guards, pattern, **annotations):
        self.cfg = cfg
        self.guards = guards
        self.pattern = self.compile_pattern(pattern)
        self.annotations = annotations

    def compile_pattern(self, pattern):
        if not pattern:
            return None
        if not pattern.startswith("/"):
            pattern = "/" + pattern
        return URLPattern(pattern)

    def match_pattern(self, path_info):
        """ Match ``path_info`` against route's ``pattern``"""
        if self.pattern is None:
            if not path_info or path_info == "/":
                return "", ()
            raise NoURLPatternMatched()
        return self.pattern.match(path_info)

    def match_guards(self, request, trace):
        """ Match ``request`` against route's ``guards`` and accumulate result
        in ``trace``
        """
        for guard in self.guards:
            trace = guard(request, trace) or trace
        return trace

    def __call__(self, request):
        """ Try to match route against ``request``

        If no route was matched the :class:`routr.exc.NoMatchFound` exception
        will be raised.

        :param request:
            :class:`webob.Request` object to match route against
        """
        path_info = self.cfg.extract_path_info(request)
        return self.match(path_info, request)

    def match(self, path_info, request):
        """ Match ``request`` against route

        return:
            trace object which accumulate ``*args`` and ``**kwargs``
        :rtype:
            :class:`.Trace`

        :raises routr.exc.NoURLPatternMatched:
            if no route was matched by URL
        :raises routr.exc.RouteGuarded:
            if route was guarded by one or more guards
        :raises routr.exc.MethodNotAllowed:
            if method isn't allowed for matched route
        """
        raise NotImplementedError()

    def reverse(self, name, *args, **kwargs):
        """ Reverse route with ``name`` using ``*args`` as pattern parameters
        and ``**kwargs`` as query string parameters

        :raises routr.exc.RouteReversalError:
            if no reversal can be computed for given arguments
        """
        raise NotImplementedError()

    def __iter__(self):
        raise NotImplementedError()

class Endpoint(Route):
    """ Endpoint route

    Associated with some object ``target`` which will be returned in case of
    successful match and a ``method`` which matches against request's method.

    Additional to :class:`.Route` params are:

    :param target:
        object to associate with route
    :param method:
        HTTP method associate with route
    :param name:
        optional name, should be provided if reversal of this route is needed,
        otherwise ``None`` is allowed
    """

    def __init__(self, target, method, name, cfg, guards, pattern,
            **annotations):
        super(Endpoint, self).__init__(cfg, guards, pattern, **annotations)
        self.target = target
        self.method = method
        self.name = name

    def match_method(self, request):
        if self.method != self.cfg.extract_method(request):
            raise MethodNotAllowed()

    def match(self, path_info, request):
        path_info, args = self.match_pattern(path_info)
        if path_info:
            raise NoURLPatternMatched()
        self.match_method(request)
        trace = self.cfg.trace(args, {}, [self])
        trace = self.match_guards(request, trace)
        return trace

    def reverse(self, name, *args, **kwargs):
        if name != self.name:
            raise RouteReversalError("no route with name '%s'" % name)
        url = self.pattern.reverse(*args) if self.pattern else "/"
        if kwargs:
            url += "?" + urlencode(kwargs)
        return url

    def __iter__(self):
        return iter([self])

    def __repr__(self):
        return "%s(target=%r, guards=%r, pattern=%r)" % (
            self.__class__.__name__, self.target, self.guards,
            self.pattern.pattern if self.pattern else None)

    __str__ = __repr__

class RouteGroup(Route):
    """ Route which represents a group of other routes

    Can have its own ``guards`` and a URL ``pattern``.

    Additional to :class:`.Route` params are:

    :param routes:
        a list of :class:`Route` objects
    """

    def __init__(self, routes, cfg, guards, pattern, **annotations):
        super(RouteGroup, self).__init__(cfg, guards, pattern, **annotations)
        self.routes = routes

    def index(self):
        """ Return mapping from route name to actual route"""
        idx = {}
        for r in self.routes:
            if isinstance(r, Endpoint) and r.name:
                if r.name in idx:
                    raise RouteConfigurationError(
                        "route this name '%s' already defined")
                if self.pattern or r.pattern:
                    idx[r.name] = self.pattern + r.pattern
                else:
                    idx[r.name] = URLPattern("/")
            elif isinstance(r, RouteGroup):
                ridx = r.index()
                if set(ridx) & set(idx):
                    raise RouteConfigurationError(
                        "route this name '%s' already defined")
                for (n, u) in ridx.items():
                    idx[n] = self.pattern + u
        return idx

    @cached_property
    def _cached_index(self):
        return self.index()

    def reverse(self, name, *args, **kwargs):
        if not name in self._cached_index:
            raise RouteReversalError("no route with name '%s'" % name)
        url = self._cached_index[name].reverse(*args)
        if kwargs:
            url += "?" + urlencode(kwargs)
        return url

    def match_pattern(self, path_info):
        """ Match ``path_info`` against route's ``pattern``"""
        if self.pattern is None:
            return path_info, ()
        return self.pattern.match(path_info)

    def match(self, path_info, request):
        path_info, args = self.match_pattern(path_info)
        guarded = []
        trace = self.cfg.trace(args, {}, [self])
        trace = self.match_guards(request, trace)
        for subroute in self.routes:
            try:
                subtrace = subroute.match(path_info, request)
            except NoURLPatternMatched:
                continue
            except MethodNotAllowed, e:
                guarded.append(RouteGuarded(e, e.response))
                continue
            except RouteGuarded, e:
                guarded.append(e)
                continue
            except HTTPException, e:
                guarded.append(RouteGuarded(e, e))
                continue
            else:
                return ((trace + subtrace)
                    if subtrace is not None and trace is not None
                    else trace or subtrace)
        if guarded:
            # NOTE we raise only last guard failure
            # cause it's more interesting one
            raise guarded[-1]
        raise NoURLPatternMatched()

    def __iter__(self):
        return iter(self.routes)

    def __repr__(self):
        return "%s(routes=%r, guards=%r, pattern=%r)" % (
            self.__class__.__name__, self.routes, self.guards, self.pattern)

    __str__ = __repr__

def parse_args(line):
    args = []
    kwargs = {}
    if not line:
        return args, kwargs
    for item in (a.strip() for a in line.split(",") if a):
        if "=" in item:
            k, v = item.split("=", 1)
            kwargs[k.strip()] = v.strip()
        else:
            args.append(item)
    return args, kwargs

def handle_str(args):
    if args:
        raise InvalidRoutePattern("'str' type doesn't accept args")
    return ("[^/]+", None)

def handle_path(args):
    if args:
        raise InvalidRoutePattern("'path' type doesn't accept args")
    return (".*", None)

def handle_int(args):
    if args:
        raise InvalidRoutePattern("'path' type doesn't accept args")
    return ("[0-9]+", int)

def handle_any(args):
    args, kwargs = parse_args(args)
    if not args:
        raise InvalidRoutePattern("'any' type requires positional args")
    if kwargs:
        raise InvalidRoutePattern("'any' doesn't accept keyword args")

    return ("(" + "|".join("(" + re.escape(x) + ")" for x in args) + ")", None)

class URLPattern(object):

    _type_re = re.compile("""
        {
        (?P<label>[a-z]+)            # label
        (:(?P<type>[a-z]+))?         # optional type identifier
        (\(                          # optional args
            (?P<args>[a-zA-Z= ,_]*)
        \))?
        }""", re.VERBOSE)


    _typemap = {
        None:       handle_str,
        "str":      handle_str,
        "string":   handle_str,
        "path":     handle_path,
        "int":      handle_int,
        "any":      handle_any,
    }

    def __init__(self, pattern):
        self.pattern = pattern

        self._compiled = None
        self._names = None

    @cached_property
    def is_exact(self):
        return self._type_re.search(self.pattern) is None

    @cached_property
    def compiled(self):
        if self._compiled is None:
            self.compile()
        return self._compiled

    @cached_property
    def _pattern_len(self):
        return len(self.pattern)

    def compile(self):
        if self.is_exact:
            return

        names = []
        compiled = ""
        last = 0
        for n, m in enumerate(self._type_re.finditer(self.pattern)):
            compiled += re.escape(self.pattern[last:m.start()])
            typ, label, args = (
                m.group("type"), m.group("label"), m.group("args"))
            if not typ in self._typemap:
                raise InvalidRoutePattern(
                    "unknown type '%s' in pattern '%s'" % (typ, self.pattern))
            r, c = self._typemap[typ](args)
            name = "_gpt%d" % n
            names.append((name, c, label))
            compiled += "(?P<%s>%s)" % (name, r)
            last = m.end()
        compiled += re.escape(self.pattern[last:])

        self._compiled = re.compile(compiled)
        self._names = names

    def reverse(self, *args):
        if self.is_exact:
            return self.pattern

        r = self.pattern
        for arg in args:
            r = self._type_re.sub(str(arg), r, 1)
        if self._type_re.search(r):
            raise RouteReversalError(
                "not enough params for reversal of '%s' route,"
                " only %r was supplied" % (self.pattern, args))
        return r

    def match(self, path_info):
        if self.is_exact:
            if not path_info.startswith(self.pattern):
                raise NoURLPatternMatched(path_info)
            return path_info[self._pattern_len:], ()

        m = self.compiled.match(path_info)
        if not m:
            raise NoURLPatternMatched("no match for '%s' against '%s'" % (
                path_info, self._compiled.pattern))
        groups = m.groupdict()
        try:
            args = tuple(
                c(groups[n]) if c else groups[n]
                for (n, c, l) in self._names)
        except ValueError:
            raise NoURLPatternMatched()
        return path_info[m.end():], args

    def __add__(self, o):
        if o is None:
            return self
        return URLPattern(join(self.pattern, o.pattern))

    def __radd__(self, o):
        if o is None:
            return self
        return URLPattern(join(o.pattern, self.pattern))

    def __repr__(self):
        return "<routr.URLPattern %s>" % self.pattern

def include(spec):
    """ Include routes by ``spec``

    :param spec:
        asset specification which points to :class:`.Route` instance
    """
    r = import_string(spec)
    if not isinstance(r, Route):
        raise RouteConfigurationError(
            "route included by '%s' isn't a route" % spec)
    return r

def plug(name):
    """ Plug routes by ``setuptools`` entry points, identified by ``name``

    :param name:
        entry point name to query routes
    """
    routes = []
    for p in iter_entry_points("routr", name=name):
        r = p.load()
        if not isinstance(r, Route):
            raise RouteConfigurationError(
                "entry point '%s' doesn't point at Route instance" % p)
        routes.append(r)
    return RouteGroup(routes, [])

class Configuration(object):
    """ Object which holds configuration for route definition DSL"""

    endpoint = Endpoint
    group = RouteGroup
    trace = Trace

    def extract_method(self, request):
        """ Extract method of the ``request``"""
        return request.method

    def extract_path_info(self, request):
        """ Extract PATH_INFO of the ``request``"""
        return request.path_info

    def route(self, *args, **kwargs):
        """ Directive for configuring routes in application

        :param args:
            ([method,] [pattern,] target) produces endpoint route
            ([method,] [pattern,] *routes) produces route group
        :param kwargs:
            name and guards are treated as name and guards for routes, other
            keyword args a are passed as annotations
        """

        def consume(pred):
            r = []
            while args:
                if not pred(args[0]):
                    break
                r.append(args.pop(0))
            return r, args

        args = list(args)

        if not args:
            raise RouteConfigurationError("empty routes")

        method = args.pop(0) if isinstance(args[0], HTTPMethod) else GET
        name = kwargs.pop("name", None)

        if not args:
            raise RouteConfigurationError("empty routes")

        if len(args) == 1 and not isinstance(args[0], Route):
            target = args[0]
            return self.endpoint(target, method, name, self, [], None, **kwargs)

        elif (
                len(args) == 2
                and isinstance(args[0], str)
                and not isinstance(args[1], Route)):
            pattern, target = args
            return self.endpoint(target, method, name, self, [], pattern,
                    **kwargs)

        else:
            if isinstance(args[0], str):
                pattern = args.pop(0)
            else:
                pattern = None

            guards, args = consume(
                lambda d: not isinstance(d, Route) and hasattr(d, "__call__"))

            routes, args = consume(
                lambda d: isinstance(d, Route))

            if routes:
                return self.group(routes, self, guards, pattern, **kwargs)
            elif len(args) == 1:
                target = args[0]
                return self.endpoint(target, method, name, self, guards,
                        pattern, **kwargs)
            else:
                raise RouteConfigurationError(
                    "improper usage of 'route' directive")

config = Configuration()
route = config.route



"""

    routr.tests -- test suite
    =========================

"""

from unittest import TestCase
from webob import Request, exc

from routr.schema import String, Int, opt, qs
from routr import Route, Endpoint, RouteGroup, URLPattern
from routr import route, RouteConfigurationError
from routr import POST, GET
from routr.exc import (
    NoURLPatternMatched, RouteGuarded, MethodNotAllowed, RouteReversalError)

__all__ = ()

class TestRouting(TestCase):

    def assertNoMatch(self, r, url):
        if isinstance(url, Request):
            self.assertRaises(NoURLPatternMatched, r, url)
        else:
            self.assertRaises(NoURLPatternMatched, r, Request.blank(url))

class TestRootEnpoint(TestRouting):

    def test_reverse(self):
        r = route("target", name="news")
        self.assertEqual(r.reverse("news"), "/")
        self.assertRaises(RouteReversalError, r.reverse, "news2")

    def test_match(self):
        r = route("target")
        req = Request.blank("/")
        tr = r(req)
        self.assertEqual(len(tr.routes), 1)
        self.assertEqual(
            (tr.args, tr.kwargs, tr.target),
            ((), {}, "target"))

    def test_no_match(self):
        def target():
            return "hello"
        r = route(target)
        req = Request.blank("/news")
        self.assertRaises(NoURLPatternMatched, r, req)

class TestEndpoint(TestRouting):

    def test_reverse(self):
        r = route("news", "target", name="news")
        self.assertEqual(r.reverse("news"), "/news")
        self.assertRaises(RouteReversalError, r.reverse, "news2")

        r = route("news/{id:int}/", "target", name="news")
        self.assertEqual(r.reverse("news", 42), "/news/42/")
        self.assertRaises(RouteReversalError, r.reverse, "news2")

        r = route("news/{any(login,attach,app_login,access_token)}",
            "target", name="news")
        self.assertEqual(r.reverse("news", "aa"), "/news/aa")
        self.assertRaises(RouteReversalError, r.reverse, "news2")

    def test_match(self):
        r = route("news", "target")
        req = Request.blank("/news")
        tr = r(req)
        self.assertEqual(len(tr.routes), 1)
        self.assertEqual(
            (tr.args, tr.kwargs, tr.target),
            ((), {}, "target"))

    def test_no_match(self):
        self.assertNoMatch(
            route("news", "target"),
            "/new")
        self.assertNoMatch(
            route("news", "target"),
            "/newsweek")

    def test_method(self):
        r = route(POST, "news", "target")
        req = Request.blank("/news", {"REQUEST_METHOD": "POST"})
        tr = r(req)
        self.assertEqual(
            (tr.args, tr.kwargs, tr.target),
            ((), {}, "target"))

        self.assertRaises(
            MethodNotAllowed,
            r, Request.blank("/news", {"REQUEST_METHOD": "DELETE"}))

    def test_param_pattern_int(self):
        r = route("/news/{id:int}/", "target")
        req = Request.blank("/news/42/")
        tr = r(req)
        self.assertEqual(
            (tr.args, tr.kwargs, tr.target),
            ((42,), {}, "target"))

        self.assertNoMatch(r, Request.blank("/news/"))
        self.assertNoMatch(r, Request.blank("/news/a/"))
        self.assertNoMatch(r, Request.blank("/news//"))
        self.assertNoMatch(r, Request.blank("/news/122"))

        r = route("/news/{a:int}/{b:int}/{c:int}/", "target")
        req = Request.blank("/news/42/41/40/")
        tr = r(req)
        self.assertEqual(
            (tr.args, tr.kwargs, tr.target),
            ((42, 41, 40), {}, "target"))

    def test_param_pattern_string(self):
        r = route("/news/{id:string}/", "target")

        req = Request.blank("/news/42/")
        tr = r(req)
        self.assertEqual(
            (tr.args, tr.kwargs, tr.target),
            ((u"42",), {}, "target"))

        req = Request.blank("/news/abcdef-12/")
        tr = r(req)
        self.assertEqual(
            (tr.args, tr.kwargs, tr.target),
            ((u"abcdef-12",), {}, "target"))

    def test_param_pattern_path(self):
        r = route("/news/{p:path}", "target")

        req = Request.blank("/news/42/news")
        tr = r(req)
        self.assertEqual(
            (tr.args, tr.kwargs, tr.target),
            ((u"42/news",), {}, "target"))

        r = route("/news/{p:path}/comments", "target")

        req = Request.blank("/news/42/news/comments")
        tr = r(req)
        self.assertEqual(
            (tr.args, tr.kwargs, tr.target),
            ((u"42/news",), {}, "target"))

    def test_param_guard(self):
        r = route(
            "/news/{id:int}/",
            qs(q=opt(String), page=opt(Int)),
            "target")

        req = Request.blank("/news/42/")
        tr = r(req)
        self.assertEqual(
            (tr.args, tr.kwargs, tr.target),
            ((42,), {}, "target"))

        req = Request.blank("/news/42/?q=search")
        tr = r(req)
        self.assertEqual(
            (tr.args, tr.kwargs.items(), tr.target),
            ((42,), [("q", "search")], "target"))

        req = Request.blank("/news/42/?q=search&page=100")
        tr = r(req)
        self.assertEqual(
            (tr.args, tr.kwargs.items(), tr.target),
            ((42,), [("q", "search"), ("page", 100)], "target"))

        self.assertRaises(
            exc.HTTPBadRequest,
            r, Request.blank("/news/42/?q=search&page=aa"))

class TestRouteGroup(TestRouting):

    def test_reverse(self):
        r = route("api",
            route("news", "news", name="news"),
            route("comments", "comments", name="comments"))
        self.assertEqual(r.reverse("news"), "/api/news")
        self.assertEqual(r.reverse("comments"), "/api/comments")
        self.assertRaises(RouteReversalError, r.reverse, "a")

        r = route("api",
            route("news/{id}/", "news", name="news"),
            route("comments", "comments", name="comments"))
        self.assertEqual(r.reverse("news", "hello"), "/api/news/hello/")

        r = route("api",
            route(GET, "news", name="get-news"),
            route(POST, "news", name="create-news"))
        self.assertEqual(r.reverse("get-news"), "/api")
        self.assertEqual(r.reverse("create-news"), "/api")

    def test_reverse_empty_pattern(self):
        r = route(route("news", name="news"))
        self.assertEqual(r.reverse("news"), "/")

    def test_simple(self):
        r = route(
            route("news", "news"),
            route("comments", "comments"))

        req = Request.blank("/news")
        tr = r(req)
        self.assertEqual(len(tr.routes), 2)
        self.assertEqual(
            (tr.args, tr.kwargs, tr.target),
            ((), {}, "news"))

        req = Request.blank("/comments")
        tr = r(req)
        self.assertEqual(
            (tr.args, tr.kwargs, tr.target),
            ((), {}, "comments"))

        self.assertNoMatch(r, "/newsweeek")
        self.assertNoMatch(r, "/ne")

    def test_group_inexact_pattern(self):
        r = route("news",
                route("{id:int}",
                    route("comments", "view")))
        req = Request.blank("/news/42/comments")
        tr = r(req)
        self.assertEqual(
                (tr.args, tr.kwargs, tr.endpoint.target),
                ((42,), {}, "view"))

        r = route("news/{id:int}",
                route("comments", "view"))
        req = Request.blank("/news/42/comments")
        tr = r(req)
        self.assertEqual(
                (tr.args, tr.kwargs, tr.endpoint.target),
                ((42,), {}, "view"))

        r = route("news",
                route("{id:int}/comments", "view"))
        req = Request.blank("/news/42/comments")
        tr = r(req)
        self.assertEqual(
                (tr.args, tr.kwargs, tr.endpoint.target),
                ((42,), {}, "view"))

    def test_complex_match(self):
        def news():
            return "news"
        def comments():
            return "comments"
        def api_news():
            return "api_news"
        def api_comments():
            return "api_comments"

        r = route(
            route("api",
                route("news", "api_news"),
                route("comments", "api_comments")),
            route("news", "news"),
            route("comments", "comments"))

        req = Request.blank("/news")
        tr = r(req)
        self.assertEqual(
            (tr.args, tr.kwargs, tr.target),
            ((), {}, "news"))

        req = Request.blank("/comments")
        tr = r(req)
        self.assertEqual(
            (tr.args, tr.kwargs, tr.target),
            ((), {}, "comments"))

        req = Request.blank("/api/news")
        tr = r(req)
        self.assertEqual(
            (tr.args, tr.kwargs, tr.target),
            ((), {}, "api_news"))

        req = Request.blank("/api/comments")
        tr = r(req)
        self.assertEqual(
            (tr.args, tr.kwargs, tr.target),
            ((), {}, "api_comments"))

    def test_by_method(self):
        r = route("api",
            route(GET, "news_get"),
            route(POST, "news_post"))

        req = Request.blank("/api", {"REQUEST_METHOD": "POST"})
        tr = r(req)
        self.assertEqual(
            (tr.args, tr.kwargs, tr.target),
            ((), {}, "news_post"))

        req = Request.blank("/api")
        tr = r(req)
        self.assertEqual(
            (tr.args, tr.kwargs, tr.target),
            ((), {}, "news_get"))

    def test_method_inner(self):
        r = route(
            route("news", "news"),
            route(GET, "comments", "comments_get"),
            route(POST, "comments", "comments_post"))

        req = Request.blank("/news", {"REQUEST_METHOD": "GET"})
        tr = r(req)
        self.assertEqual(
            (tr.args, tr.kwargs, tr.target),
            ((), {}, "news"))

        req = Request.blank("/news", {"REQUEST_METHOD": "POST"})
        self.assertRaises(
            RouteGuarded,
            r, req)

        req = Request.blank("/comments", {"REQUEST_METHOD": "POST"})
        tr = r(req)
        self.assertEqual(
            (tr.args, tr.kwargs, tr.target),
            ((), {}, "comments_post"))

        req = Request.blank("/comments", {"REQUEST_METHOD": "DELETE"})
        self.assertRaises(
            RouteGuarded,
            r, req)

        self.assertNoMatch(r, "/newsweeek")
        self.assertNoMatch(r, "/ne")

    def test_guards(self):
        pass

class TestRouteDirective(TestCase):

    def test_root_endpoint(self):
        r = route("myapp.mytarget")
        self.assertEqual(r.pattern, None)
        self.assertEqual(r.guards, [])
        self.assertIsInstance(r, Endpoint)
        self.assertEqual(r.target, "myapp.mytarget")

    def test_root_endpoint_func(self):
        def target():
            pass
        r = route(target)
        self.assertEqual(r.pattern, None)
        self.assertEqual(r.guards, [])
        self.assertIsInstance(r, Endpoint)
        self.assertEqual(r.target, target)

    def test_root_endpoint_guards(self):
        r = route(id, "myapp.mytarget")
        self.assertEqual(r.pattern, None)
        self.assertEqual(r.guards, [id])
        self.assertIsInstance(r, Endpoint)
        self.assertEqual(r.target, "myapp.mytarget")

    def test_endpoint(self):
        r = route("news", "myapp.mytarget")
        self.assertNotEqual(r.pattern, None)
        self.assertEqual(r.guards, [])
        self.assertIsInstance(r, Endpoint)
        self.assertEqual(r.target, "myapp.mytarget")

    def test_endpoint_func(self):
        def target():
            pass
        r = route("news", target)
        self.assertNotEqual(r.pattern, None)
        self.assertEqual(r.guards, [])
        self.assertIsInstance(r, Endpoint)
        self.assertEqual(r.target, target)

    def test_endpoint_guards(self):
        r = route("news", id, "myapp.mytarget")
        self.assertNotEqual(r.pattern, None)
        self.assertEqual(r.guards, [id])
        self.assertIsInstance(r, Endpoint)
        self.assertEqual(r.target, "myapp.mytarget")

    def test_route_list_no_pattern(self):
        r = route(
            route("news", "myapp.api.news"),
            route("comments", "myapp.api.comments"))
        self.assertEqual(r.guards, [])
        self.assertEqual(r.pattern, None)
        self.assertIsInstance(r, RouteGroup)

    def test_route_list_no_pattern_guards(self):
        r = route(
            id,
            route("news", "myapp.api.news"),
            route("comments", "myapp.api.comments"))
        self.assertEqual(r.guards, [id])
        self.assertEqual(r.pattern, None)
        self.assertIsInstance(r, RouteGroup)

    def test_route_list(self):
        r = route("api",
            route("news", "myapp.api.news"),
            route("comments", "myapp.api.comments"))
        self.assertEqual(r.guards, [])
        self.assertNotEqual(r.pattern, None)
        self.assertIsInstance(r, RouteGroup)

    def test_route_list_guards(self):
        r = route("api",
            id,
            route("news", "myapp.api.news"),
            route("comments", "myapp.api.comments"))
        self.assertNotEqual(r.pattern, None)
        self.assertEqual(r.guards, [id])
        self.assertIsInstance(r, RouteGroup)

    def test_invalid_routes(self):
        self.assertRaises(RouteConfigurationError, route)

class TestURLPattern(TestCase):

    def test_int(self):
        p = URLPattern("/a/{id:int}/b/")
        self.assertTrue(not p.is_exact)
        self.assertEqual(p.match("/a/42/b/"), ('', (42,)))

    def test_str(self):
        p = URLPattern("/a/{id}/b/")
        self.assertTrue(not p.is_exact)
        self.assertEqual(p.match("/a/42/b/"), ('', ("42",)))

        p = URLPattern("/a/{id:str}/b/")
        self.assertTrue(not p.is_exact)
        self.assertEqual(p.match("/a/42/b/"), ('', ("42",)))

        p = URLPattern("/a/{id:string}/b/")
        self.assertTrue(not p.is_exact)
        self.assertEqual(p.match("/a/42/b/"), ('', ("42",)))

    def test_path(self):
        p = URLPattern("/a/{id:path}/b/")
        self.assertTrue(not p.is_exact)
        self.assertEqual(p.match("/a/42/43/b/"), ('', ("42/43",)))

    def test_any(self):
        p = URLPattern("/a/{id:any(aaa, bbb, ccc)}/b/")
        self.assertTrue(not p.is_exact)
        self.assertEqual(p.match("/a/aaa/b/"), ('', ("aaa",)))
        self.assertEqual(p.match("/a/bbb/b/"), ('', ("bbb",)))
        self.assertEqual(p.match("/a/ccc/b/"), ('', ("ccc",)))
        self.assertRaises(NoURLPatternMatched, p.match, "/a")
        self.assertRaises(NoURLPatternMatched, p.match, "/a/abc/b/")



"""

    wk.doc -- extension for generating documentation with sphinx
    ============================================================

"""

import sys
from fnmatch import fnmatchcase

from docutils import nodes
from docutils.statemachine import ViewList

from sphinx.util.compat import Directive
from sphinx.util.docstrings import prepare_docstring
from sphinx.util.nodes import nested_parse_with_titles
from sphinxcontrib import httpdomain

from routr.utils import import_string, cached_property, join
from routr import RouteGroup

__all__ = ("AutoRoutrDirective",)

def traverse_routes(route, path=None):
    """ Traverse routes by flatten them"""

    def acc_path(acc, r):
        if not r.pattern:
            return acc
        return join(acc, r.pattern.pattern)

    path = path or (route.pattern.pattern if route.pattern else None) or "/"
    if isinstance(route, RouteGroup):
        return [(m, p, r)
            for subroute in route.routes
            for (m, p, r) in traverse_routes(
                subroute,
                path=acc_path(path, subroute))]
    else:
        return [(route.method, path, route)]

def http_directive(method, path, content):
    """ Construct line for http directive from httpdomain

    :copyright: (c) Hong Minhee
    """
    method = method.lower().strip()
    if isinstance(content, basestring):
        content = content.splitlines()
    yield ""
    yield ".. http:%s:: %s" % (method, path)
    yield ""
    for line in content:
        yield "   " + line
    yield ""

class AutoRoutrDirective(Directive):
    """ Directive for generating docs from routr routes

    Based on code from httpdomain.autoflask by Hong Minhee.
    """

    has_content = True
    requied_argument = 1
    option_spec = {
        "include": str,
        "exclude": str,
    }

    @cached_property
    def include_patterns(self):
        if "include" in self.options:
            return [x.strip() for x in self.options["include"].split(",")]
        return []

    @cached_property
    def exclude_patterns(self):
        if "exclude" in self.options:
            return [x.strip() for x in self.options["exclude"].split(",")]
        return []

    def allowed(self, path):
        return (
            (not self.include_patterns
            or any(fnmatchcase(path, x) for x in self.include_patterns))
                and
            (not self.exclude_patterns
            or all(not fnmatchcase(path, x) for x in self.exclude_patterns)))

    def make_rst(self):
        routes = import_string(self.content.data[0])
        for method, path, route in traverse_routes(routes):
            if not self.allowed(path):
                continue
            if isinstance(route.target, basestring):
                try:
                    target = import_string(route.target)
                except ImportError:
                    print >> sys.stderr, "cannot import %s" % route.target
                    continue
            else:
                target = route.target
            docstring = prepare_docstring(target.__doc__ or "")
            for line in http_directive(method, path, docstring):
                yield line

    def run(self):
        node = nodes.section()
        node.document = self.state.document
        result = ViewList()
        for line in self.make_rst():
            result.append(line, "<autoroutr>")
        nested_parse_with_titles(self.state, result, node)
        return node.children

def setup(app):
    if "http" not in app.domains:
        httpdomain.setup(app)
    app.add_directive("autoroutr", AutoRoutrDirective)



"""

    routr.schema -- constrain routes
    ================================

"""

from webob import exc
from colander import * # pylint: disable=W0401,W0614,W0622

__all__ = (
    "RequestParams", "FormParams", "QueryParams", "Optional",
    "opt", "qs", "form")

_none = object()

class Optional(object):

    none = _none

    def __init__(self, typ, default=_none):
        if isinstance(typ, type):
            typ = typ()
        self.typ = typ
        self.default = default

class RequestParams(object):
    """ Guard for query string parameters

    Raises :class:`webob.exc.HTTPBadRequest` if query string isn't validated.

    :param kwargs:
        mapping with validators for query string
    """

    exception_factory = exc.HTTPBadRequest

    def __init__(self, **kwargs):
        self.schema = SchemaNode(Mapping())
        self.post_processor = None
        for name, typ in kwargs.items():
            if isinstance(typ, SchemaNode):
                typ = typ.clone()
                typ.name = name
                self.schema.add(typ)
                continue
            if isinstance(typ, Optional):
                missing = typ.default
                typ = typ.typ
                if isinstance(typ, SchemaNode):
                    typ = typ.clone()
                    typ.name = name
                    typ.missing = missing
                    self.schema.add(typ)
                else:
                    self.schema.add(SchemaNode(
                        typ,
                        name=name,
                        missing=missing))
            else:
                if isinstance(typ, type):
                    typ = typ()
                self.schema.add(SchemaNode(typ, name=name))

    def params(self, request):
        raise NotImplementedError()

    def then(self, post_processor):
        self.post_processor = post_processor
        return self

    def __call__(self, request, trace):
        try:
            kwargs = self.schema.deserialize(self.params(request))
        except Invalid, e:
            raise self.exception_factory(e)
        for k, v in kwargs.items():
            if v is Optional.none:
                kwargs.pop(k)
        if self.post_processor:
            kwargs = self.post_processor(kwargs)
        trace.kwargs.update(kwargs)
        return trace

    def __add__(self, o):
        s = SchemaNode(Mapping())
        for c in self.schema.children:
            s.add(c)
        for c in o.schema.children:
            s.add(c)
        r = QueryParams()
        r.schema = s
        if self.post_processor or o.post_processor:
            def post_processor(kwargs):
                ret_kwargs = {}
                if self.post_processor:
                    ret_kwargs.update(self.post_processor(kwargs))
                if o.post_processor:
                    ret_kwargs.update(o.post_processor(kwargs))
                return ret_kwargs
            r.post_processor = post_processor
        return r

class QueryParams(RequestParams):

    def params(self, request):
        return request.GET

class FormParams(RequestParams):

    def params(self, request):
        return request.POST

qs = QueryParams
form = FormParams
opt = Optional



from setuptools import setup, find_packages
import sys, os

version = "0.2"

setup(
    name="routr",
    version=version,
    description="URL routing made right",
    long_description=open("README").read() + "\n\n" + open("CHANGES").read(),
    author="Andrey Popp",
    author_email="8mayday@gmail.com",
    license="BSD",
    packages=find_packages(exclude=["ez_setup", "examples", "tests"]),
    install_requires=[
        "colander >= 0.9.5",
        "WebOb >= 1.2b3"
    ],
    include_package_data=True,
    test_suite="routr.tests",
    zip_safe=False)
